import torch
import torch.nn as nn
import torch.optim as optim
from torchvision import datasets, transforms
from torch.utils.data import DataLoader

# （1）数据准备
# 定义数据转换
transform = transforms.Compose([
    transforms.ToTensor(),  # 将图片转换为Tensor
    transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))  # 归一化
])

# 下载CIFAR10数据集
trainset = datasets.CIFAR10(root='./data', train=True,
                            download=True, transform=transform)
trainloader = DataLoader(trainset, batch_size=64,
                         shuffle=True, num_workers=2)

testset = datasets.CIFAR10(root='./data', train=False,
                           download=True, transform=transform)
testloader = DataLoader(testset, batch_size=64,
                        shuffle=False, num_workers=2)


# （3）搭建CNN模型
class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)  # 假设输入图片被压缩到5x5
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        x = self.pool(torch.relu(self.conv1(x)))
        x = self.pool(torch.relu(self.conv2(x)))
        x = x.view(-1, 16 * 5 * 5)  # 展平
        x = torch.relu(self.fc1(x))
        x = torch.relu(self.fc2(x))
        x = self.fc3(x)
        return x

def train():
    net = Net()

    # （4）训练模型
    criterion = nn.CrossEntropyLoss()
    optimizer = optim.SGD(net.parameters(), lr=0.01, momentum=0.9)

    for epoch in range(10):  # 遍历数据集多次
        running_loss = 0.0
        for i, data in enumerate(trainloader, 0):
            inputs, labels = data

            optimizer.zero_grad()

            outputs = net(inputs)
            loss = criterion(outputs, labels)
            loss.backward()
            optimizer.step()

            running_loss += loss.item()
            if i % 2000 == 1999:  # 每2000个mini-batches打印一次
                print(f'[{epoch + 1}, {i + 1:5d}] loss: {running_loss / 2000:.3f}')
                running_loss = 0.0

    print('Finished Training')

    # （5）保存训练好的模型参数
    torch.save(net.state_dict(), 'cifar_net.pth')
    # （6）加载训练好的模型，在测试集上进行测试，并输出在测试集上的准确率

    # 加载模型
    net.load_state_dict(torch.load('cifar_net.pth'))
    net.eval()  # 设置为评估模式

    correct = 0
    total = 0
    with torch.no_grad():  # 在评估模式下不需要计算梯度
        for data in testloader:
            images, labels = data
            outputs = net(images)
            _, predicted = torch.max(outputs.data, 1)
            total += labels.size(0)
            correct += (predicted == labels).sum().item()

    print(f'Accuracy of the network on the 10000 test images: {100 * correct / total}%')

    # （7）选取测试集中第15张图片，并输出真实标签以及预测标签

    # 注意：因为testloader的迭代器会耗尽，所以我们需要重新创建或重置它
    # 一种简单的方法是再次加载测试集，但只取第一张（因为索引是从0开始的，所以第15张图片的索引是14）

    # 重新加载测试集（或者你可以找到一种方法来重置DataLoader的迭代器）
    testset = datasets.CIFAR10(root='./data', train=False, download=False, transform=transform)
    testloader_single = DataLoader(testset, batch_size=1, shuffle=False)

    # 遍历测试集直到找到第15张图片
    for i, (images, labels) in enumerate(testloader_single):
        if i == 14:  # 因为索引从0开始
            break

            # 获取真实标签和预测标签
    _, true_label = torch.max(labels, 0)
    _, predicted_label = torch.max(net(images), 1)

    print(f'True Label: {true_label.item()}, Predicted Label: {predicted_label.item()}')


if __name__ == '__main__':
    train()